#include <string.h>
#include <stdlib.h>
#include <time.h>
#include "ui.h"
#include "board.h"
#include "main.h"
#include "search.h"

int checker_pos[100];
extern int num_moves;

int board2checkers(int in)
{
    return checker_pos[in] + 1;
}
int checkers2board(int in)
{
    return board_pos[in - 1];
}
void print_move(char *comment, struct ch_move *move, int nr)
{
    int i, j = move->hops ? move->hops : 1;

    ui_chatmsg(comment);
    ui_chatmsg(" %d is: ", nr);
    for (i = 0; i < j; i++) 
	ui_chatmsg("%d-", board2checkers (move->moveto[i]));
    ui_chatmsg("%d\n", board2checkers (move->moveto[i]));
} 

void dump_ml (struct ch_movelist *ml)
{
    int i, j;

    for (i = 0; i < ml->num; i++) {
	for (j = 0; j <= ml->moves[i].hops; j++) 
	    ui_chatmsg("%d-", ml->moves[i].moveto[j]);
	ui_chatmsg("\n");
    }
}

int do_take(struct ch_board *b_in, int *moves, int mc)
{
    struct ch_movelist *ml;
    int i, j, t, c = 0;
    struct ch_move *move = 0;

    ml = move_generate (b_in, IS_BLACK(b_in->square[moves[0]]));

    if (!ml->num)
	return 0;

    for (i = 0; i < ml->num; i++) {
	if (ml->moves[i].moveto[0] != moves[0]) 
	    continue;
	t = 0;
	for (j = 0; j <= ml->moves[i].hops; j++) {
	    if (ml->moves[i].moveto[j] == moves[t])
		t++;
	}
	if (t == mc) {
	    move = &ml->moves[i];
	    c++;
	}
    }

    if (c > 1 || !move) {
	if (move) ui_chatmsg("Move is ambigous!");
	else ui_chatmsg("No possibilities found!\n");
	free(ml);
	return 0;
    } else {
	move_make (b_in, b_in, move);
	free(ml);
	return 1;
    }
}

int do_move(struct ch_board *b_in, int *moves, int mc) 
{
    int a = abs(moves[1] - moves[0]);

    if (mc == 2) {
	if (a != 11 && a != 9)
	    return do_take(b_in, moves, mc);
	if (b_in->square[moves[1]] != B_EMPTY) {
	    ui_chatmsg("Destination not empty!\n");
	    return 0;
	}
	/* check legality of move */
	if (move_cantake (b_in, IS_BLACK(b_in->square[moves[0]]))) {
	    ui_chatmsg ("You must jump!\n");
	    return 0;
	}
	    
	if (IS_BLACK(b_in->square[moves[0]])) {
	    if (!(moves[0] + 11 == moves[1] || moves[0] + 9 == moves[1]))
		if (b_in->square[moves[0]] == B_BKING 
		    && !(moves[0] - 11 == moves[1]||moves[0] - 9 == moves[1]))
		    return 0;
	} else if (IS_WHITE(b_in->square[moves[0]])) {
	    if (!(moves[0] - 11 == moves[1] || moves[0] - 9 == moves[1]))
		if (b_in->square[moves[0]] == B_BKING
                    && !(moves[0] + 11 == moves[1]||moves[0] + 9 == moves[1]))
		    return 0;
	}

	/* Make the move */
	b_in->square[moves[1]] = b_in->square[moves[0]];
	b_in->square[moves[0]] = B_EMPTY;
	move_promote (b_in, moves[1]);
    } else
	return do_take(b_in, moves, mc);

    return 1;
}

int parse_move(struct ch_board *b_in, char *ms)
{
    int j, n = 0, mc = 0;
    int moves[32];
    char *tok;

    tok = strtok(ms, "-");
    j = checkers2board(atoi(tok));
    if (b_in->isblack) {
	if (!IS_WHITE(b_in->square[j]))
	    return 0;
    } else if (!IS_BLACK(b_in->square[j]))
	return 0;

    while (tok) {
	n++;
	j = checkers2board(atoi(tok));
	moves[mc++] = j;
	tok = strtok(0, "-");
    }

    free(ms);
    if (n > 1) {
	if (!do_move(b_in, moves, mc))
	    return 0;
	return 1;
    } else 
	return 0;
}

void clean_exit(int i)
{
    ui_shutdown();
    exit(i);
}

void init_game()
{
    int x;

    ui_init();
    srand(time(0));

    for (x = 0; x < 100; x++)
	checker_pos[x] = 0;
    for (x = 0; x < 32; x++)
	checker_pos[board_pos[x]] = x;

    search_init();
    move_init();
}

void move_test(struct ch_board *b_in)
{
    struct ch_movelist *ml;
    int startclock = clock();

    while (clock() < startclock + 10 * CLOCKS_PER_SEC) {
	ml = move_generate(b_in, 0);
	free(ml);
    }

    ui_chatmsg("In 10 sec generated %d moves\n", num_moves);
    ui_getmsg("quit?");
    clean_exit(0);
	    
}

void self_play (struct ch_board *board)
{
    struct ch_move *move;
    float btime = 0, wtime = 0;
    int tmp;
    
    board->isblack = TRUE;

    for (;;) {
	tmp = board->my_moves;
	board->my_moves = board->op_moves;
	board->op_moves = tmp;

	if (board->isblack) {
	    wtime = board->elapsed_time;
	    board->elapsed_time = btime;
	} else {
	    btime = board->elapsed_time;
	    board->elapsed_time = wtime;
	}

	move = search_move(board);
	if (board->isblack)
	    print_move("Black", move, ++board->my_moves);
	else
	    print_move("White", move, ++board->my_moves);
	move_make(board, board, move);
	debug("Eval computer: %d, util %d",
	      board_eval(board, board->isblack), move->util);
	ui_update(board);
	board->isblack = !board->isblack;

	if (board_iswin(board, FALSE)) {
	    ui_getmsg("\nWhite has won!\n Bang on <enter> to quit");
	    clean_exit(0);
	} else if (board_iswin(board, FALSE)) {
	    ui_getmsg("\nBlack has won!\n Bang on <enter> to quit");
	    clean_exit(0);
	}
    }
}

struct ch_move *first_move(struct ch_board *b_in, char *msg)
{
    static struct ch_move first;

    debug("First move %s", msg);
    
    if (b_in->isblack) {
	first.hops = 0;
	first.moveto[0] = 36;
	first.moveto[1] = 45;
	return &first;
    } else {
	if (!strcmp(msg, "11-15")) {
	    first.hops = 0;
	    first.moveto[0] = 67;
	    first.moveto[1] = 58;
	    return &first;
	} else if (!strcmp(msg, "11-16")) {
	    first.hops = 0;
            first.moveto[0] = 63;
            first.moveto[1] = 54;
	    return &first;
        } else if (!strcmp(msg, "12-16")) {
	    first.hops = 0;
            first.moveto[0] = 67;
            first.moveto[1] = 56;
	    return &first;
	} else if (!strcmp(msg, "10-15")) {
	    first.hops = 0;
            first.moveto[0] = 61;
            first.moveto[1] = 52;
	    return &first;
	} else if (!strcmp(msg, "10-14")) {
	    first.hops = 0;
            first.moveto[0] = 63;
            first.moveto[1] = 52;
	    return &first;
        } else if (!strcmp(msg, "9-14")) {
	    first.hops = 0;
            first.moveto[0] = 67;
            first.moveto[1] = 58;
            return &first;
	} else if (!strcmp(msg, "9-13")) {
	    first.hops = 0;
            first.moveto[0] = 63;
            first.moveto[1] = 54;
            return &first;
	}
    }
    return 0;
}

int main()
{
    struct ch_board *board = board_init(), *bt;
    struct ch_move *move;
    struct stack *stackski;
    char *msg;
    int i = 0, r = 0;

    init_game();

    stackski = stack_init();
    board->total_time = 300;
    ui_update(board);

    msg = ui_getmsg("Am i b/w? ");

    if (!strcmp(msg, "w")) {
	ui_chatmsg("Computer playing white\n");
	board->isblack = 0;
	i = 1;
    } else if (!strcmp(msg, "b")) {
	ui_chatmsg("Computer playing black\n");
        board->isblack = 1;
    } else if (!strcmp(msg, "mt"))
	move_test(board);
    else {
	ui_chatmsg("Selfplay!\n");
	self_play(board);
    }
   
    for(;;) {
	if (!i) {
	    if (board->my_moves || !(move = first_move (board, msg)))	       
		move = search_move(board);
	    if (move) {
		print_move("My move", move, ++board->my_moves);
		move_make(board, board, move);
		debug("Eval computer: %d, util %d", 
		      board_eval(board, board->isblack), move->util);
		ui_update(board);
	    }
	    if (!move || board_iswin(board, board->isblack)) {
		ui_getmsg("\nI won!!!!!.\n Bang on <enter> to quit");
		clean_exit(0);
	    }
	}
	
	i = 0;
	do {
	    r = 0;
	    msg = ui_getmsg("Enter move: ");
	    if (strlen(msg) < 3) 
		r = 1;
	    else if (!strcmp(msg, "back")) {
		if (stackski->count) {
		    ui_chatmsg("Backed one move\n");
		    board = (struct ch_board *)pop(stackski);
		    ui_update(board);
		} else 
		    ui_chatmsg("No moves!");
		r = 1;
	    }
	    bt = malloc(sizeof(struct ch_board));
	    memcpy (bt, board, sizeof(struct ch_board));
	} while (r || !parse_move(board, strdup(msg)));
	push(stackski, bt);
	board->op_moves++;
	debug("Eval opponent: %d", board_eval(board, !board->isblack));
	ui_update(board);
	if (board_iswin(board, !board->isblack)) {
	    ui_getmsg("\nOpponent has won.\n Bang on <enter> to quit");
	    clean_exit(0);
	}
    }
    
    ui_getmsg("Pound <enter> to exit");
    ui_shutdown();

    return 0;
}


