#include <stdlib.h>
#include <limits.h>
#include "main.h"
#include "board.h"
#include "search.h"

static struct ch_board boards[MAXMOVES];
int history[32][32], killers = -1;
struct ch_move killer[MAXDEPTH];
extern int checker_pos[100];

#define B_CANJUMP(i, n) (IS_WHITE(b_in->square[i + n]) && \
                          b_in->square[i + 2*n] ==  B_EMPTY)
#define W_CANJUMP(i, n) (IS_BLACK(b_in->square[i + n]) && \
                          b_in->square[i + 2*n] ==  B_EMPTY)
#define TAKE_MOVE(n) { move_jump(ml, i, i + n, fork, hops); \
                   b_work = &boards[hops]; \
                   board_clone(b_in, b_work); \
                   b_work->square[i + n] = b_work->square[i]; \
                   b_work->square[i + (n/2)] = B_EMPTY; \
                   move_take (ml, b_work, i + n, black, hops + 1); \
                   fork = 1; }

#define NOTAKE_MOVE(n) { ml->moves[ml->num].hops = 0; \
                     ml->moves[ml->num].hash = 18*i + n; \
                     ml->moves[ml->num].moveto[0] = i; \
		     ml->moves[ml->num].moveto[1] = i + n; \
                     num_moves++;\
                     ml->num++; }

int num_moves = 0;

void move_init()
{
    int i, x, y;

    for (i = 0; i < MAXMOVES; i++)
        board_clear(&boards[i]);

    debug("Clearing history");
    for (x = 0; x < 32; x++)
	for (y = 0; y < 32; y++)
	    history[x][y] = 0;
}

void inline move_clone(struct ch_move *ms, struct ch_move *md)
{
    memcpy(md, ms, sizeof(struct ch_move));
}

void move_clearkill()
{
    memset(killer, 0, sizeof(struct ch_move) * MAXDEPTH);
    killers = -1;
}

void move_killer(struct ch_move *move, int ply) 
{
    if (killers < ply || move->util > killer[ply].util) 
	move_clone(move, &killer[ply]);
    killers = ply;
}

struct ch_move *move_findkill(struct ch_movelist *ml, int ply)
{
    int i, h;

    if (killers < ply)
	return 0;

    for (i = 0; i < ml->num; i++)
	if (ml->moves[i].hash == killer[ply].hash 
	    && ml->moves[i].hops == killer[ply].hops) {
	    h = killer[ply].hops ? killer[ply].hops + 1: 2;
	    if (!memcmp(ml->moves[i].moveto, killer[ply].moveto, 
			h * sizeof(int)))
		return &ml->moves[i];
	}

    return 0;
}

void move_history(struct ch_move *move)
{
    int x, y;

    x = checker_pos[move->moveto[0]];
    y = move->hops ? checker_pos[move->moveto[move->hops]]
        : checker_pos[move->moveto[1]];
    history[x][y]++;
}

void move_tofront(struct ch_movelist *ml)
{
    int b, i, x, y;
    struct ch_move *move;

    ml->hist = move = &ml->moves[0];
    x = checker_pos[move->moveto[0]];
    y = move->hops ? checker_pos[move->moveto[move->hops]] 
	: checker_pos[move->moveto[1]];
    b = history[x][y];
    for (i = 1; i < ml->num; i++) {
	move = &ml->moves[i];
	x = checker_pos[move->moveto[0]];
	y = move->hops ? checker_pos[move->moveto[move->hops]]
	    : checker_pos[move->moveto[1]];
	if (history[x][y] > b) {
	    ml->hist = move;
	    b = history[x][y];
	}
    }
}

void move_promote(struct ch_board *b_in, int i)
{
    /* Promote to kings */
    if (IS_BLACK(b_in->square[i]) && i > 80 && i < 89)
	b_in->square[i] = B_BKING;
    else if (IS_WHITE(b_in->square[i]) && i > 10 && i < 19)
	b_in->square[i] = B_WKING;
}

static inline struct ch_movelist *move_listinit()
{
    struct ch_movelist *ml = malloc(sizeof(struct ch_movelist));
    ml->num = 0;

    return ml;
}

void move_make(struct ch_board *b_in, struct ch_board *b_out, 
	       struct ch_move *move)
{
    int i = 1, piece = 324;

    board_clone(b_in, b_out);
    b_out->elapsed_time += move->time;

    if (move->hops == 0) {
	b_out->square[move->moveto[1]] = b_out->square[move->moveto[0]];
	b_out->square[move->moveto[0]] = B_EMPTY;
    } else {
	piece = b_out->square[move->moveto[0]];
	b_out->square[move->moveto[0]] = B_EMPTY;
	for(i = 0; i < move->hops; i++) {
	    if (move->moveto[i] + 22 == move->moveto[i + 1]) 
		b_out->square[move->moveto[i] + 11] = B_EMPTY;
	    else if (move->moveto[i] + 18 == move->moveto[i + 1]) 
		b_out->square[move->moveto[i] + 9] = B_EMPTY;
	    else if (move->moveto[i] - 22 == move->moveto[i + 1]) 
		b_out->square[move->moveto[i] - 11] = B_EMPTY;
	    else if (move->moveto[i] - 18 == move->moveto[i + 1]) 
		b_out->square[move->moveto[i] - 9] = B_EMPTY;
	}
	b_out->square[move->moveto[i]] = piece;
    }
#ifdef DEBUG
    if (piece == B_EMPTY) {
	debug("Oy! move %d to %d had empty piece hops %d", 
	      move->moveto[0], move->moveto[i], move->hops);
	ui_getmsg("oy!");
	for (i = 0; i <= move->hops; i++)
	    debug("moveto[%d] = %d", i, move->moveto[i]);
	ui_update(b_in);
	ui_getmsg("Done?");
	ui_update(b_out);
    }
#endif
    move_promote(b_out, move->moveto[i]);
}

static inline int move_istake(struct ch_board *b_in, int i, int black)
{
    int *bd = b_in->square;

    if (!black) {
	if (IS_BLACK(bd[i - 11]) && bd[i - 22] == B_EMPTY) return TRUE;
	if (IS_BLACK(bd[i -  9]) && bd[i - 18] == B_EMPTY) return TRUE;
	if (bd[i] == B_WKING) {
	    if (IS_BLACK(bd[i + 11]) && bd[i + 22] == B_EMPTY) return TRUE;
	    if (IS_BLACK(bd[i +  9]) && bd[i + 18] == B_EMPTY) return TRUE;
	}
    } else {
	if (IS_WHITE(bd[i + 11]) && bd[i + 22] == B_EMPTY) return TRUE;
	if (IS_WHITE(bd[i +  9]) && bd[i + 18] == B_EMPTY) return TRUE;
	if (bd[i] == B_BKING) {
	    if (IS_WHITE(bd[i - 11]) && bd[i - 22] == B_EMPTY) return TRUE;
	    if (IS_WHITE(bd[i -  9]) && bd[i - 18] == B_EMPTY) return TRUE;
	}
    }
    return FALSE;
}

int move_cantake(struct ch_board *b_in, int black)
{
   int i;

   for (i = 0; i < 32; i++)
       if (b_in->square[board_pos[i]] > B_EMPTY && 
	   IS_BLACK(b_in->square[board_pos[i]]) == black
	   && move_istake(b_in, board_pos[i], black))
	   return TRUE;
   return FALSE;
}

static void move_notake(struct ch_movelist *ml, struct ch_board *b_in, 
		      int i)
{
    if (b_in->square[i] == B_BLACK) {
	if (b_in->square[i + 9] == B_EMPTY) NOTAKE_MOVE (9);
	if (b_in->square[i + 11] == B_EMPTY) NOTAKE_MOVE (11);
    } else if (b_in->square[i] == B_WHITE) {
	if (b_in->square[i - 9] == B_EMPTY) NOTAKE_MOVE (-9);
	if (b_in->square[i - 11] == B_EMPTY) NOTAKE_MOVE (-11);
    } else { // Is a king
	if (b_in->square[i + 9] == B_EMPTY) NOTAKE_MOVE (9);
	if (b_in->square[i + 11] == B_EMPTY) NOTAKE_MOVE (11);
	if (b_in->square[i - 9] == B_EMPTY) NOTAKE_MOVE (-9);
	if (b_in->square[i - 11] == B_EMPTY) NOTAKE_MOVE (-11);
    }
}

static inline void move_jump(struct ch_movelist *ml, int i, int j, int fork, 
			     int hops)
{
    struct ch_move *move;

    if (fork) {
	move_clone(&ml->moves[ml->num], &ml->moves[ml->num + 1]);
	ml->num++;
    }
  
    move = &ml->moves[ml->num];
    move->hops = hops;
    move->moveto[move->hops++] = i;
    move->moveto[move->hops] = j;
    move->hash = move->hash * 17 + j;

#ifdef DEBUG
    if ((i + 22) != j &&
	(i - 22) != j &&
	(i + 18) != j &&
	(i - 18) != j)
	debug("movebug: i %d, j %d, hops %d, fork %d", i, j, hops, fork);
#endif
}

static void move_take(struct ch_movelist *ml, struct ch_board *b_in, int i, 
		      int black, int hops)
{
    int fork = 0;
    struct ch_board *b_work;

    /* Every recursion needs a stopping-condition. This is our */
    if (!move_istake(b_in, i, black)) {
	ml->moves[ml->num].num = num_moves++;
	return;
    }

    if (b_in->square[i] == B_BLACK) {
	if (B_CANJUMP(i, 9)) TAKE_MOVE (18);
	if (B_CANJUMP(i, 11)) TAKE_MOVE (22);
    } else if (b_in->square[i] == B_WHITE) {
	if (W_CANJUMP(i, -9)) TAKE_MOVE (-18);
	if (W_CANJUMP(i, -11)) TAKE_MOVE (-22); 
    } else if (b_in->square[i] == B_BKING) {
	if (B_CANJUMP(i, 9)) TAKE_MOVE (18);
	if (B_CANJUMP(i, 11)) TAKE_MOVE (22);
	if (B_CANJUMP(i, -9)) TAKE_MOVE (-18);
	if (B_CANJUMP(i, -11)) TAKE_MOVE (-22);
    } else {
	if (W_CANJUMP(i, 9)) TAKE_MOVE (18);
	if (W_CANJUMP(i, 11)) TAKE_MOVE (22);
	if (W_CANJUMP(i, -9)) TAKE_MOVE (-18);
	if (W_CANJUMP(i, -11)) TAKE_MOVE (-22);
    }
}

struct ch_movelist *move_generate(struct ch_board *b_in, int black)
{
    int i, s;
    struct ch_movelist *takes, *notakes;

    takes = move_listinit();
    notakes = move_listinit();

    if (black) {
	for (i = 0; i < 32; i++) {
	    s = board_pos[i];
	    if (b_in->square[s] > B_EMPTY 
		&& IS_BLACK(b_in->square[s]) == black) {
		if (move_istake(b_in, s, black)) {
		    move_take(takes, b_in, s, black, 0);
		    takes->num++;
		} else if (!takes->num)
		    move_notake(notakes, b_in, s);
	    }
	}
    } else {
	for (i = 31; i > -1 ; i--) {
            s = board_pos[i];
            if (b_in->square[s] > B_EMPTY
                && IS_BLACK(b_in->square[s]) == black) {
                if (move_istake(b_in, s, black)) {
                    move_take(takes, b_in, s, black, 0);
                    takes->num++;
                } else if (!takes->num)
                    move_notake(notakes, b_in, s);
            }
        }
    }
    if (takes->num) {
	move_tofront(takes);
	move_freeall(notakes);
	return takes;
    } else {
	move_tofront(notakes);
	move_freeall(takes);
	return notakes;
    }
}

void move_freeall(struct ch_movelist *ml)
{
    if (!ml)
	return;
    free(ml);
}


