#include <stdlib.h>
#include <time.h>
#include <limits.h>
#include "main.h"
#include "board.h"
#include "search.h"

extern int num_moves;
extern int checker_pos[100];
int num_nodes = 0, num_evals = 0, num_prune = 0;
static struct ch_board boards[MAXDEPTH + 2];

int maxply;

void search_init()
{
    int i;

    for (i = 0; i < MAXDEPTH + 2; i++)
	board_clear(&boards[i]);
}

float ch_allowedtime(struct ch_board *b_in)
{
    if (b_in->my_moves < 8)
	return 9;
    return b_in->my_moves < 40 ? 
	(b_in->total_time - b_in->elapsed_time) / (50 - b_in->my_moves) 
	: (b_in->total_time - b_in->elapsed_time) / (70 - b_in->my_moves);
}

static int search_ab(struct ch_board *b_in, int ply, int maxclock, 
		       int alpha, int beta, int black)
{
    struct ch_move *move = 0, *bestmove = 0, *killer = 0;
    struct ch_movelist *ml;
    struct ch_board *b_work = &boards[ply];
    int testbeta = beta;
    int i = 0;

    num_nodes++;

    /* Nice one. Stopping condition on clock and maxply.
     * for maxply also considers Quiescence-heuristic */
    if ((clock() >= maxclock)
	|| ((ply >= maxply) && !move_cantake(b_in, black))) {
	num_evals++;
#ifdef DEBUG
	if (ply == MAXDEPTH)
	    debug("Hit maxdepth!");
#endif DEBUG
	return b_in->isblack == black ? board_eval(b_in, black) 
	    : -board_eval(b_in, black);
    }

    /* Generate all moves for current player and descend */
    ml = move_generate(b_in, black);

    if (ml->num) {
	move = ml->hist;
	killer = move_findkill(ml, ply);
    }
    while (move) {
	move_make(b_in, b_work, move);
	move->util = -search_ab(b_work, ply + 1, maxclock,
			     -beta, -alpha, !black);
	if (alpha < move->util) {
	    bestmove = move;
	    alpha = move->util;
	    testbeta = alpha + 1;
	    if (alpha >= beta) {
		num_prune++;
		i = ml->num;
		break;
	    }
	}
	if (move == ml->hist && move != killer)
	    move = killer; 
	else
	    move = 0;		
    }

    for (; i < ml->num; i++) {
	move = &ml->moves[i];
	if (move != ml->hist && move != killer) {
	    move_make(b_in, b_work, move);
	    move->util = -search_ab(b_work, ply + 1, maxclock, 
				 -testbeta, -alpha, !black);
	    if (move->util >= testbeta && testbeta < beta)
		move->util = -search_ab(b_work, ply + 1, maxclock,
				     -beta, -alpha, !black);
	    if (alpha < move->util) {
		alpha = move->util;
		testbeta = alpha + 1;
		bestmove = move;
		if (alpha >= beta) {
		    num_prune++;
		    i = ml->num;
		}
	    }
	}
    }
    if (bestmove) {
	move_history(bestmove);
	move_killer(bestmove, ply);
    }
    i = ml->num;
    move_freeall(ml);

    if (i) 
	return alpha;
     else
	return b_in->isblack == black ? board_eval(b_in, black) 
	    : -board_eval(b_in, black);
}

struct ch_move *search_pre(struct ch_board *b_in, struct ch_movelist *ml, 
			   int alpha, int beta, int maxclock, int ply)
{
    int i, testbeta = beta;
    struct ch_move *bestmove = 0, *move, *killer = 0;
    struct ch_board *b_work = &boards[ply];

    move = ml->hist;
    killer = move_findkill(ml, ply);
    while (move) {
	move_make (b_in, b_work, move);
	move->util = -search_ab(b_work, ply + 1, maxclock, -beta, -alpha,
				!b_in->isblack);
	if (move->util > alpha) {
	    alpha = move->util;
	    bestmove = move;
	    testbeta = alpha + 1;
	}
	if (move == ml->hist && move != killer)
	    move = killer; 
	else
	    move = 0;		
    }

    for (i = 0; i < ml->num; i++) {	
	move = &ml->moves[i];
	if (move != ml->hist && move != killer) {
	    move_make (b_in, b_work, move);
	    move->util = -search_ab(b_work, ply + 1, maxclock, -testbeta, 
				    -alpha, !b_in->isblack);
	    if (move->util >=testbeta && testbeta < beta)
		move->util = -search_ab(b_work, ply + 1, maxclock, -beta, 
					-alpha, !b_in->isblack);
	    if (move->util > alpha) {
		alpha = move->util;
		bestmove = move;
		testbeta = alpha + 1;
	    }
	}
    }
    return bestmove;
}

struct ch_move *search_move(struct ch_board *b_in)
{
    float availtime;
    int maxclock, startclock;
    struct ch_movelist *ml;
    struct ch_move *move, *result, *tmpres = 0;

    result = malloc (sizeof(struct ch_move));

    /* Clear statistic variables */
    num_nodes = 0;
    num_evals = 0;
    num_moves = 0;
    num_prune = 0;

    maxply = STARTDEPTH;
    move_clearkill();
    
    /* How much time do we have this time? */
    availtime = ch_allowedtime(b_in);
    startclock = clock();
    maxclock = startclock + CLOCKS_PER_SEC * availtime;
    ui_infomsg("Available time: %f, maxclock %d\n", availtime, maxclock);

    /* Generate the moves we can make, and take care of
     * the base cases */
    ml = move_generate(b_in, b_in->isblack);
    if (!ml->num)
	return 0; /* No possible moves */
    if (ml->num == 1) /* Only one move, don't search! */
	tmpres = &ml->moves[0];
    else { 
	/* This is the interresting part... */
	while (clock() < maxclock && maxply < MAXDEPTH) {
	    move_tofront(ml);
	    move = search_pre (b_in, ml, -INF, INF, maxclock, 0);
	    if (!move) {
		debug("No move returned!");
		return 0;
	    }
	    debug("Finished %d plies, time %d", maxply, clock() - startclock);
	    maxply++;
	    /* If we have a new candidate, update our result */
	    if (clock() < maxclock) {
		tmpres = move;
		move_history (tmpres);
		move_killer (tmpres, 0);
	    }
	}
	debug("Search finished. Max depth reached %d", maxply - 2);
    }

    memcpy (result, tmpres, sizeof (struct ch_move));
    debug ("Move utility %d", result->util);
    
    ui_infomsg("Generated %d moves\n", num_moves);
    ui_infomsg("Visited %d nodes\n", num_nodes);
    ui_infomsg("Evaluated %d leafs\n", num_evals);
    ui_infomsg("Pruned %d subtrees\n", num_prune);

    result->time = (float)(clock() - startclock) / (float)CLOCKS_PER_SEC;
    return result;
}








