#include <stdio.h>
#include <curses.h>
#include <time.h>
#include "main.h"
#include "ui.h"
#include "board.h"

static WINDOW *board, *info, *chat;
extern int checker_pos[100];

/* Initialize the UI routines. Must be done before everything */
void ui_init()
{
    int x_max, y_max;

    initscr();

    /* Get screen parameters and see if we can work with this. */
    getmaxyx(stdscr, y_max, x_max);
    if (y_max < 13 || x_max < 40) {
	endwin();
	puts("Cannot use screen, wrong size");
	exit(0);
    }
    clear();

    /* Setup our windows */
    board = newwin(11, 11, 0, 0);
    info = newwin(11, x_max - 13, 0, 13);
    chat = newwin(y_max - 13, x_max, 12, 0);

    if (!(board && info && chat)) {
	endwin();
	puts("Couldn't initialize windows");
	exit(0);
    }

    /* Set scrolling attributes */
    scrollok(chat, TRUE);
    idlok(chat, TRUE);
    scrollok(info, TRUE);
    idlok(info, TRUE);

    /* Partition the workspace */
    mvhline(11, 0, ACS_HLINE, x_max);
    mvhline(y_max - 2, 0, ACS_HLINE, x_max);
    mvvline(0, 12, ACS_VLINE, 11);
    mvaddch(11, 12, ACS_BTEE);

    ui_update(0);
}

/* Shutdown. Done before exiting the program */
void ui_shutdown()
{
    delwin(board);
    delwin(info);
    delwin(chat);
    endwin();
}
void ui_chatmsg(char *s, ...)
{
    static char msg[1024];
    va_list alist;
    
    va_start(alist, s);
    
    vsnprintf(msg, 1024, s, alist);

    waddstr(chat, msg);
    wrefresh(chat);

    va_end(alist);
}

void ui_infomsg(char *s, ...)
{
    static char msg[1024];
    va_list alist;
    
    va_start(alist, s);
    
    vsnprintf(msg, 1024, s, alist);

    waddstr(info, msg);
    wrefresh(info);

    va_end(alist);
}

/* WARNING: The string returned here is STATIC! strdup it if you need it
 * elsewhere! */
char *ui_getmsg(char *out)
{
    static char msg[128];

    waddstr(chat, out);
    wrefresh(info);
    wgetnstr(chat, msg, 120);

    return msg;
}

/* Update some of the attributes of the ui */
void ui_update(struct ch_board *b_in)
{
    int x_max, y_max, piece;
    char buf[256];
    int x, y;
    
    if (!b_in) 
	return;

    sprintf(buf, "CPU-Time used: %f secs  Moves Total: %d, My: %d", 
	    b_in->elapsed_time, b_in->op_moves+b_in->my_moves,
	    b_in->my_moves);
    getmaxyx(stdscr, y_max, x_max);
    mvaddstr(y_max - 1, 0, buf);
    refresh();
    
    mvwaddstr(board, 0, 2, "12345678");
    for (y = 2; y < 10; y++)
	mvwaddch(board, y, 0, '0'+y-1);

    /* Update gameboard */
    for (y = 0; y < 10; y++)
	for (x = 0; x < 10; x++) {
	    piece = b_in->square[y * 10 + x];
	    if (piece == B_WHITE) piece = 'w';
	    else if (piece == B_WKING) piece = 'W';
	    else if (piece == B_BLACK) piece = 'b';
	    else if (piece == B_BKING) piece = 'B';
	    else if (piece == B_INVAL) piece = '*';
	    else if (piece == B_EMPTY) piece = '.';
	    else piece = '!';
	    if ((y * 10 + x) == 12 || checker_pos[y * 10 + x] || piece == '*')
		mvwaddch(board, y+1, x+1, piece);
	}
    wrefresh(board);
}

